#include <stdio.h>
/* malloc, abs */
#include <stdlib.h>

#include "array.h"
#include "grid-detail.h"

#include "util-grid.h"

static int grid_is_open(int c);

static int
grid_is_open(int c)
{
  switch (c)
  {
  case GR_FLOOR_ROOM:
  case GR_FLOOR_CORRIDOR:
    return 1;
    break;
  default:
    break;
  }

  return 0;
}

int
dump_grid(int **grid, int size_x, int size_y)
{
  int c;
  int x;
  int y;

  if (grid == NULL)
  {
    fprintf(stderr, "dump_grid: grid is NULL\n");
    return 1;
  }
  if (size_x <= 0)
  {
    fprintf(stderr, "dump_grid: size_x is non-positive (%d)\n", size_x);
    return 1;
  }
  if (size_y <= 0)
  {
    fprintf(stderr, "dump_grid: size_y is non-positive (%d)\n", size_y);
    return 1;
  }

  for (y = 0; y < size_y; y++)
  {
    for (x = 0; x < size_x; x++)
    {
      switch (grid[x][y])
      {
      case GR_FLOOR_ROOM:
        c = '.';
        break;
      case GR_FLOOR_CORRIDOR:
        c = ':';
        break;
      case GR_WALL:
        c = '#';
        break;
      default:
        c = '?';
        break;
      }
      putchar(c);
    }
    putchar('\n');
  }

  return 0;
}

int
fill_grid_with_wall(int **grid, int size_x, int size_y)
{
  int x;
  int y;

  if (grid == NULL)
  {
    fprintf(stderr, "fill_grid_with_wall: grid is NULL\n");
    return 1;
  }
  if (size_x <= 0)
  {
    fprintf(stderr, "fill_grid_with_wall: size_x is non-positive (%d)\n",
            size_x);
    return 1;
  }
  if (size_y <= 0)
  {
    fprintf(stderr, "fill_grid_with_wall: size_y is non-positive (%d)\n",
            size_y);
    return 1;
  }

  for (x = 0; x < size_x; x++)
  {
    for (y = 0; y < size_y; y++)
    {
      grid[x][y] = GR_WALL;
    }
  }

  return 0;
}

int
draw_line(int **grid, int size_x, int size_y,
          int ax, int ay, int bx, int by,
          int value, int thick)
{
  int u;
  int v;
  /* r = 2 * target_v * (v - v_real_line)
   * note that r can be negative; r must be signed
   */
  int r;
  int x0;
  int y0;
  int x1;
  int y1;
  int dx;
  int dy;
  int dx_abs;
  int dy_abs;
  int dir;
  int temp;
  int target_u;
  int target_v;

  if (grid == NULL)
  {
    fprintf(stderr, "draw_line: grid is NULL\n");
    return 1;
  }
  if (size_x <= 0)
  {
    fprintf(stderr, "draw_line: size_x is non-positive (%d)\n", size_x);
    return 1;
  }
  if (size_y <= 0)
  {
    fprintf(stderr, "draw_line: size_y is non-positive (%d)\n", size_y);
    return 1;
  }
  if ((ax == bx) && (ay == by))
  {
    fprintf(stderr, "draw_line: two points are same (%d, %d)\n", ax, ay);
    return 1;
  }

  dx = bx - ax;
  dy = by - ay;
  dx_abs = abs(dx);
  dy_abs = abs(dy);

  if (dx >= 0)
  {
    if (dy >= 0)
    {
      if (dx_abs >= dy_abs)
        dir = 0;
      else
        dir = 1;
    }
    else
    {
      if (dx_abs >= dy_abs)
        dir = 7;
      else
        dir = 6;
    }
  }
  else
  {
    if (dy >= 0)
    {
      if (dx_abs >= dy_abs)
        dir = 3;
      else
        dir = 2;
    }
    else
    {
      if (dx_abs >= dy_abs)
        dir = 4;
      else
        dir = 5;
    }
  }

  if (dx_abs >= dy_abs)
  {
    target_u = dx_abs;
    target_v = dy_abs;
  }
  else
  {
    target_u = dy_abs;
    target_v = dx_abs;
  }


  /* the Bresenham algorithm */
  v = 0;
  r = 0;
  for (u = 0; u <= target_u; u++)
  {
    x0 = u;
    y0 = v;
    x1 = u;
    y1 = v + 1;

    if ((dir & 1) == 1)
    {
      temp = x0;
      x0 = y0;
      y0 = temp;

      temp = x1;
      x1 = y1;
      y1 = temp;
    }
    if ((dir & 2) == 2)
    {
      temp = x0;
      x0 = -y0;
      y0 = temp;

      temp = x1;
      x1 = -y1;
      y1 = temp;
    }
    if ((dir & 4) == 4)
    {
      x0 = -x0;
      y0 = -y0;

      x1 = -x1;
      y1 = -y1;
    }

    x0 += ax;
    y0 += ay;
    x1 += ax;
    y1 += ay;

    /* if v_new - v_real_line <= (1 / 2) * (target_v / target_u) */
    if (r >= target_u)
    {
      if ((x1 >= 0) && (x1 < size_x)
          && (y1 >= 0) && (y1 < size_y))
      { 
        grid[x1][y1] = value;
      }
      if (thick)
      {
        if ((x0 >= 0) && (x0 < size_x)
            && (y0 >= 0) && (y0 < size_y))
        { 
          grid[x0][y0] = value;
        }
      }

      r -= 2 * target_u;
      v++;
    }
    else
    {
      if ((x0 >= 0) && (x0 < size_x)
          && (y0 >= 0) && (y0 < size_y))
      { 
        grid[x0][y0] = value;
      }
    }

    r += 2 * target_v;
  }

  return 0;
}

int
check_connection(int **grid, int size_x, int size_y)
{
  int x;
  int y;
  int x0;
  int y0;
  int dx;
  int dy;
  int found;
  int num_frontier;
  int *frontier_x = NULL;
  int *frontier_y = NULL;
  int **flood = NULL;

  if (grid == NULL)
  {
    fprintf(stderr, "check_connection: grid is NULL\n");
    return 1;
  }
  if (size_x <= 0)
  {
    fprintf(stderr, "check_connection: size_x is non-positive (%d)\n",
            size_x);
    return 1;
  }
  if (size_y <= 0)
  {
    fprintf(stderr, "check_connection: size_y is non-positive (%d)\n",
            size_y);
    return 1;
  }

  frontier_x = (int *) malloc(sizeof(int) * size_x * size_y);
  if (frontier_x == NULL)
  {
    fprintf(stderr, "check_connection: malloc(frontier_x) failed\n");
    return 1;
  }
  frontier_y = (int *) malloc(sizeof(int) * size_x * size_y);
  if (frontier_y == NULL)
  {
    fprintf(stderr, "check_connection: malloc(frontier_y) failed\n");
    free(frontier_x);
    frontier_x = NULL;
    return 1;
  }
  flood = array2_new(size_x, size_y);
  if (flood == NULL)
  {
    fprintf(stderr, "check_connection: array2_new() failed\n");
    free(frontier_x);
    frontier_x = NULL;
    free(frontier_y);
    frontier_y = NULL;
    return 1;
  }

  for (x = 0; x < size_x; x++)
  {
    for (y = 0; y < size_y; y++)
    {
      flood[x][y] = 0;
    }
  }

  num_frontier = 0;
  found = 0;
  for (x = 0; x < size_x; x++)
  {
    if (found)
      break;
    for (y = 0; y < size_y; y++)
    {
      if (!grid_is_open(grid[x][y]))
        continue;
      found = 1;
      frontier_x[0] = x;
      frontier_y[0] = y;
      num_frontier = 1;
      flood[x][y] = 1;
      break;
    }
  }

  if (!found)
  {
    fprintf(stderr, "check_connection: the map has no open grid\n");
    free(frontier_x);
    frontier_x = NULL;
    free(frontier_y);
    frontier_y = NULL;
    array2_delete(flood, size_x, size_y);
    flood = NULL;
    return 1;
  }

  while (num_frontier > 0)
  {
    x0 = frontier_x[num_frontier - 1];
    y0 = frontier_y[num_frontier - 1];
    num_frontier--;
    for (dx = -1; dx <= 1; dx++)
    {
      for (dy = -1; dy <= 1; dy++)
      {
        x = x0 + dx;
        y = y0 + dy;

        if ((dx == 0) && (dy == 0))
          continue;
        if ((x < 0) || (x >= size_x))
          continue;
        if ((y < 0) || (y >= size_y))
          continue;

        if (!grid_is_open(grid[x][y]))
          continue;
        if (flood[x][y] != 0)
          continue;

        flood[x][y] = 1;
        frontier_x[num_frontier] = x;
        frontier_y[num_frontier] = y;
        num_frontier++;
      }
    }
  }

  found = 0;
  for (x = 0; x < size_x; x++)
  {
    if (found)
      break;
    for (y = 0; y < size_y; y++)
    {
      if (!grid_is_open(grid[x][y]))
        continue;
      if (flood[x][y] == 0)
      {
        fprintf(stderr, "check_connection: the map is not connected\n");
        found = 1;
        break;
      }
    }
  }

  free(frontier_x);
  frontier_x = NULL;
  free(frontier_y);
  frontier_y = NULL;
  array2_delete(flood, size_x, size_y);
  flood = NULL;

  return found;
}
